#include <iostream>
#include <vector>

using std::cin;
using std::cout;
using std::cerr;
using std::endl;
using std::vector;

enum content_types
{
  nothing,
  urinium,
  arsetope,
  coolant,
  ihd,
  irp,
  water_bucket,
  ice,
  num_content_types
};

struct cell
{
  bool functioning();
  bool failed();

  unsigned int type_of_contents;
  unsigned int amount_of_contents;
  int stored_heat;
  int life_remaining;
};

bool cell::functioning()
  {
    if(stored_heat >= 10000)
    {
      return false;
    }
    if(type_of_contents == urinium && life_remaining <= 0)
    {
      return false;
    }
    if(type_of_contents == water_bucket && amount_of_contents <= 0)
    {
      return false;
    }
    if(type_of_contents == ice && amount_of_contents <= 0)
    {
      return false;
    }
    return true;
  }

bool cell::failed()
  {
    if(type_of_contents == nothing)
    {
      return false;
    }
    if(stored_heat >= 10000)
    {
      return true;
    }
    return false;
  }

struct reactor
{
  vector<vector<cell> > cells;
  int hull_stored_heat;
  int external_cooling;

  int external_buckets_accumulated;
  int partial_external_bucket;
  int external_bucket_pumps;
  int max_external_buckets;

  int external_ice_accumulated;
  int partial_external_ice;
  int external_ice_pumps;

  int ewes_generated;

  int peak_hull_stored_heat;
  int ticks_evaluated;

  static const int ewe_per_pulse= 10;

  void reset_stats();

  bool evaulate_tick();
  void urinium_eval(size_t x, size_t y);
  void arsetope_eval(size_t x, size_t y);
  void coolant_eval(size_t x, size_t y);
  void ihd_eval(size_t x, size_t y);
  void irp_eval(size_t x, size_t y);
  void bucket_eval(size_t x, size_t y);
  void ice_eval(size_t x, size_t y);
};

void reactor::reset_stats()
  {
    hull_stored_heat= 0;
    peak_hull_stored_heat= 0;
    ticks_evaluated= 0;
    external_buckets_accumulated= 0;
    partial_external_bucket= 0;
    external_ice_accumulated= 0;
    partial_external_ice= 0;
    ewes_generated= 0;
    for(size_t x= 0; x < cells.size(); ++x)
    {
      for(size_t y= 0; y < cells[x].size(); ++y)
      {
        cells[x][y].amount_of_contents= 1;
        cells[x][y].stored_heat= 0;
        cells[x][y].life_remaining= 10000;
        if(cells[x][y].type_of_contents == ice)
        {
          cells[x][y].amount_of_contents= 64;
        }
      }
    }
  }

bool reactor::evaulate_tick()
  {
    bool keep_evaluating= false;

    for(size_t x= 0; x < cells.size(); ++x)
    {
      for(size_t y= 0; y < cells[x].size(); ++y)
      {
        if(!cells[x][y].functioning())
        {
          continue;
        }
        switch(cells[x][y].type_of_contents)
        {
          case urinium:
            urinium_eval(x, y);
            if(cells[x][y].functioning())
            {
              keep_evaluating= true;
            }
            break;
          case arsetope:
            arsetope_eval(x, y);
            break;
          case coolant:
            coolant_eval(x, y);
            break;
          case ihd:
            ihd_eval(x, y);
            break;
          case irp:
            irp_eval(x, y);
            break;
          case water_bucket:
            bucket_eval(x, y);
            break;
          case ice:
            ice_eval(x, y);
            break;
          case nothing:
          default:
            break;
        }
      }
    }
    if(hull_stored_heat > 0)
    {
      hull_stored_heat+= external_cooling;
    }
    if(hull_stored_heat < 0)
    {
      hull_stored_heat= 0;
    }
    if(hull_stored_heat>= 4000 && external_buckets_accumulated > 0)
    {
      int num_buckets_to_use= ((hull_stored_heat - 4000) / 500) + 1;
      if(num_buckets_to_use > external_buckets_accumulated)
      {
        num_buckets_to_use = external_buckets_accumulated;
      }
      //cout << "Hull heat reached " << hull_stored_heat << ", using "
      //  << num_buckets_to_use << " buckets out of "
      //  << external_buckets_accumulated;
      hull_stored_heat-= num_buckets_to_use * 500;
      external_buckets_accumulated-= num_buckets_to_use;
      //cout << ".  Heat now " << hull_stored_heat << endl;
    }
    partial_external_bucket+= external_bucket_pumps;
    if(partial_external_bucket > 10)
    {
      external_buckets_accumulated+= partial_external_bucket / 10;
      partial_external_bucket%= 10;
    }
    if(external_buckets_accumulated > max_external_buckets)
    {
      external_buckets_accumulated= max_external_buckets;
    }
    if(hull_stored_heat >= 200 && external_ice_accumulated > 0)
    {
      int num_ice_to_use= ((hull_stored_heat - 200) / 200) + 1;
      if(num_ice_to_use > external_ice_accumulated)
      {
        num_ice_to_use= external_ice_accumulated;
      }
      hull_stored_heat-= num_ice_to_use * 200;
      external_ice_accumulated-= num_ice_to_use;
    }
    partial_external_ice+= external_ice_pumps;
    if(partial_external_ice > 10)
    {
      external_ice_accumulated+= partial_external_ice / 10;
      partial_external_ice%= 10;
    }
    if(hull_stored_heat > peak_hull_stored_heat)
    {
      peak_hull_stored_heat= hull_stored_heat;
    }
    ++ticks_evaluated;
    return keep_evaluating;
  }

void reactor::urinium_eval(size_t x, size_t y)
  {
    int times_to_pulse= 1;
    int times_to_heat= 1;
    vector<cell*> neighbors_to_heat;
    const int num_neighbors= 4;
    int x_changes[num_neighbors]= {-1, 0, 1, 0};
    int y_changes[num_neighbors]= {0, -1, 0, 1};
    for(int n= 0; n < num_neighbors; ++n)
    {
      size_t neighbor_x= x + x_changes[n];
      size_t neighbor_y= y + y_changes[n];
      if(neighbor_x < cells.size())
      {
        if(neighbor_y < cells[x].size())
        {
          if(cells[neighbor_x][neighbor_y].functioning())
          {
            switch(cells[neighbor_x][neighbor_y].type_of_contents)
            {
              case urinium:
                ++times_to_pulse;
                // Both need to add heat, only urinium adds power.
              case arsetope:
                ++times_to_heat;
                break;
              case coolant:
              case ihd:
              case irp:
                neighbors_to_heat.push_back(&(cells[neighbor_x][neighbor_y]));
                break;
              case nothing:
              default:
                break;
            }
          }
        }
      }
    }
    ewes_generated+= times_to_pulse * ewe_per_pulse;
    switch(neighbors_to_heat.size())
    {
      case 0:
        hull_stored_heat+= times_to_heat * 10;
        break;
      case 1:
        neighbors_to_heat[0]->stored_heat+= times_to_heat * 10;
        break;
      case 2:
        neighbors_to_heat[0]->stored_heat+= times_to_heat * 4;
        neighbors_to_heat[1]->stored_heat+= times_to_heat * 4;
        break;
      case 3:
        neighbors_to_heat[0]->stored_heat+= times_to_heat * 2;
        neighbors_to_heat[1]->stored_heat+= times_to_heat * 2;
        neighbors_to_heat[2]->stored_heat+= times_to_heat * 2;
        break;
      case 4:
        neighbors_to_heat[0]->stored_heat+= times_to_heat * 1;
        neighbors_to_heat[1]->stored_heat+= times_to_heat * 1;
        neighbors_to_heat[2]->stored_heat+= times_to_heat * 1;
        neighbors_to_heat[3]->stored_heat+= times_to_heat * 1;
        break;
    }
    --cells[x][y].life_remaining;
  }

void reactor::arsetope_eval(size_t x, size_t y)
  {
    ++hull_stored_heat;
  }

void reactor::coolant_eval(size_t x, size_t y)
  {
    if(cells[x][y].stored_heat > 0)
    {
      --cells[x][y].stored_heat;
    }
  }

void reactor::ihd_eval(size_t x, size_t y)
  {
    vector<cell*> neighbors_to_heat;
    const int num_neighbors= 4;
    int x_changes[num_neighbors]= {-1, 0, 1, 0};
    int y_changes[num_neighbors]= {0, -1, 0, 1};
    for(int n= 0; n < num_neighbors; ++n)
    {
      size_t neighbor_x= x + x_changes[n];
      size_t neighbor_y= y + y_changes[n];
      if(neighbor_x < cells.size())
      {
        if(neighbor_y < cells[x].size())
        {
          if(cells[neighbor_x][neighbor_y].functioning())
          {
            switch(cells[neighbor_x][neighbor_y].type_of_contents)
            {
              case coolant:
              case ihd:
              case irp:
                neighbors_to_heat.push_back(&(cells[neighbor_x][neighbor_y]));
                break;
              default:
                break;
            }
          }
        }
      }
    }
    for(size_t n= 0; n < neighbors_to_heat.size(); ++n)
    {
      //cerr << "IHD at " << x << ", " << y << " has " << neighbors_to_heat.size()
      //<< " neighbors and temp " << cells[x][y].stored_heat << endl;
      int exchange_amount= 0;
      exchange_amount= (neighbors_to_heat[n]->stored_heat - cells[x][y].stored_heat) / 2;
      if(exchange_amount < -6)
      {
        exchange_amount= -6;
      }
      if(exchange_amount > 6)
      {
        exchange_amount= 6;
      }
      //cerr << "Neighbor " << n << " has " << neighbors_to_heat[n]->stored_heat
      //  << " heat, trading " << exchange_amount << endl;
      neighbors_to_heat[n]->stored_heat-= exchange_amount;
      cells[x][y].stored_heat+= exchange_amount;
    }
    int hull_exchange_amount= 0;
    hull_exchange_amount= (hull_stored_heat - cells[x][y].stored_heat) / 2;
    if(hull_exchange_amount < -25)
    {
      hull_exchange_amount= -25;
    }
    if(hull_exchange_amount > 25)
    {
      hull_exchange_amount= 25;
    }
    //cerr << "IHD has " << cells[x][y].stored_heat << " heat, hull has " << hull_stored_heat << " heat, exchanging "
    //  << hull_exchange_amount << endl;
    hull_stored_heat-= hull_exchange_amount;
    cells[x][y].stored_heat+= hull_exchange_amount;
  }

void reactor::irp_eval(size_t x, size_t y)
  {
    
  }

void reactor::bucket_eval(size_t x, size_t y)
  {
    if(hull_stored_heat >= 4000)
    {
      hull_stored_heat-= 500;
      cells[x][y].amount_of_contents= 0;
    }
  }

void reactor::ice_eval(size_t x, size_t y)
  {
    if(hull_stored_heat >= 200)
    {
      hull_stored_heat-= 200;
      --cells[x][y].amount_of_contents;
    }
  }

int main()
  {
    char const * const layout[6]=
    {
      /*
      "dcccdcccd",
      "cudcccduc",
      "duccdccud",
      "cudcccduc",
      "dcccdcccd",
      "ccdcccdcc"
      */
      /*
      "cdccdcccd",
      "  ddccdcc",
      "uuucdcccd",
      "uuu ccdcc",
      "uuu dcccd",
      "uuu ccdcc"
      */
      "         ",
      "         ",
      "uuu      ",
      "uuu      ",
      "uuu      ",
      "uuu      "
    };
    reactor loki;
    loki.cells.resize(9);
    for(int x= 0; x < 9; ++x)
    {
      loki.cells[x].resize(6);
    }
    for(size_t x= 0; x < loki.cells.size(); ++x)
    {
      for(size_t y= 0; y < loki.cells[x].size(); ++y)
      {
        switch(layout[y][x])
        {
          case ' ':
            loki.cells[x][y].type_of_contents= nothing;
            break;
          case 'u':
            loki.cells[x][y].type_of_contents= urinium;
            break;
          case 'a':
            loki.cells[x][y].type_of_contents= arsetope;
            break;
          case 'c':
            loki.cells[x][y].type_of_contents= coolant;
            break;
          case 'd':
            loki.cells[x][y].type_of_contents= ihd;
            break;
          case 'w':
            loki.cells[x][y].type_of_contents= water_bucket;
            break;
          case 'i':
            loki.cells[x][y].type_of_contents= ice;
            break;
          default:
            break;
        }
      }
    }
    loki.external_cooling= -33;
    loki.external_bucket_pumps= 4;
    loki.max_external_buckets= 4;
    loki.external_ice_pumps= 13;
    loki.reset_stats();

    bool keep_evaluating= true;
    while(keep_evaluating)
    {
      keep_evaluating= loki.evaulate_tick();
      //cerr << "Tick " << loki.ticks_evaluated << " temp: "
      //  << loki.hull_stored_heat << endl;
    }

    cout << loki.ticks_evaluated << " ticks." << endl;
    int failed_components= 0;
    for(size_t x= 0; x < loki.cells.size(); ++x)
    {
      for(size_t y= 0; y < loki.cells[x].size(); ++y)
      {
        if(loki.cells[x][y].failed())
        {
          ++failed_components;
          cout << "Failed component at " << x << ", " << y << endl;
        }
      }
    }
    cout << "Peak temp: " << loki.peak_hull_stored_heat << endl;
    cout << "Ewes generated: " << loki.ewes_generated << endl;

    return 0;
  }
